from pandac.PandaModules import GeomVertexFormat, GeomVertexData, GeomVertexWriter, Geom, GeomTristrips, GeomNode, VBase4
from fft.map import Map

def coords_to_panda(x, y, z):
    return (x, z, -y)

def uv_to_panda(polygon, u, v):
    page = polygon.texture_page
    u = u / 255.0
    v = 1.0 - (page * 0.25 + v / 1023.0)
    return (u, v)


class Polygon(object):
    def __init__(self, parent):
        self.parent = parent
        self.source = None
        self.format = GeomVertexFormat.getV3n3c4t2()
        self.node_path = None
        self.is_hovered = False

    def __del__(self):
        self.node_path.remove()

    def init_node_path(self, polygon):
        self.source = polygon
        vdata = GeomVertexData('name_me', self.format, Geom.UHStatic)
        vertex = GeomVertexWriter(vdata, 'vertex')
        normal = GeomVertexWriter(vdata, 'normal')
        color = GeomVertexWriter(vdata, 'color')
        texcoord = GeomVertexWriter(vdata, 'texcoord')
        primitive = GeomTristrips(Geom.UHStatic)
        vertex.addData3f(*coords_to_panda(*polygon.A.point.coords))
        vertex.addData3f(*coords_to_panda(*polygon.B.point.coords))
        vertex.addData3f(*coords_to_panda(*polygon.C.point.coords))
        if hasattr(polygon, 'D'):
            vertex.addData3f(*coords_to_panda(*polygon.D.point.coords))
        else:
            vertex.addData3f(*coords_to_panda(*polygon.C.point.coords))
        if polygon.A.normal:
            normal.addData3f(*coords_to_panda(*polygon.A.normal.coords))
            normal.addData3f(*coords_to_panda(*polygon.B.normal.coords))
            normal.addData3f(*coords_to_panda(*polygon.C.normal.coords))
            if hasattr(polygon, 'D'):
                normal.addData3f(*coords_to_panda(*polygon.D.normal.coords))
        gray = 1.0
        color.addData4f(gray, gray, gray, 1.0)
        color.addData4f(gray, gray, gray, 1.0)
        color.addData4f(gray, gray, gray, 1.0)
        if hasattr(polygon, 'D'):
            color.addData4f(gray, gray, gray, 1.0)
        if polygon.A.texcoord:
            texcoord_A = uv_to_panda(polygon, *polygon.A.texcoord.coords)
            texcoord_B = uv_to_panda(polygon, *polygon.B.texcoord.coords)
            texcoord_C = uv_to_panda(polygon, *polygon.C.texcoord.coords)
            texcoord.addData2f(*texcoord_A)
            texcoord.addData2f(*texcoord_B)
            texcoord.addData2f(*texcoord_C)
            if hasattr(polygon, 'D'):
                texcoord_D = uv_to_panda(polygon, *polygon.D.texcoord.coords)
                texcoord.addData2f(*texcoord_D)
        primitive.addNextVertices(4)
        primitive.closePrimitive()
        geom = Geom(vdata)
        geom.addPrimitive(primitive)
        node = GeomNode('gnode')
        node.addGeom(geom)
        self.node_path = self.parent.node_path_mesh.attachNewNode(node)

    def hover(self):
        self.is_hovered = True
        self.node_path.reparentTo(self.parent.node_path_ui)
        self.node_path.setColor(0.5, 0.5, 1.0, 1.0)

    def unhover(self):
        self.is_hovered = False
        self.node_path.reparentTo(self.parent.node_path_mesh)
        self.node_path.setColor(1.0, 1.0, 1.0, 1.0)


class Ambient_Light(object):
    def __init__(self, parent):
        self.parent = parent
        self.node_path = None

    def __del__(self):
        self.parent.node_path_mesh.clearLight(self.node_path)
        self.node_path.remove()

    def init_node_path(self, light):
        from pandac.PandaModules import AmbientLight
        alight = AmbientLight('alight')
        alight.setColor(VBase4(*light.color + (1.0,)))
        self.node_path = self.parent.node_path_mesh.attachNewNode(alight)
        self.parent.node_path_mesh.setLight(self.node_path)


class Directional_Light(object):
    def __init__(self, parent):
        self.parent = parent
        self.node_path = None

    def __del__(self):
        self.parent.node_path_mesh.clearLight(self.node_path)
        self.node_path.remove()

    def init_node_path(self, light):
        from pandac.PandaModules import DirectionalLight
        dlight = DirectionalLight('dlight')
        dlight.setColor(VBase4(*light.color + (1.0,)))
        self.node_path = self.parent.node_path_mesh.attachNewNode(dlight)
        self.node_path.setPos(*coords_to_panda(*light.direction.coords))
        self.node_path.lookAt(0,0,0)
        self.parent.node_path_mesh.setLight(self.node_path)


class Background(object):
    def __init__(self, parent):
        self.parent = parent
        self.format = GeomVertexFormat.getV3c4()
        self.node_path = None

    def __del__(self):
        self.node_path.remove()

    def init_node_path(self, background_data):
        self.source = background_data
        vdata = GeomVertexData('name_me', self.format, Geom.UHStatic)
        vertex = GeomVertexWriter(vdata, 'vertex')
        color = GeomVertexWriter(vdata, 'color')
        primitive = GeomTristrips(Geom.UHStatic)
        film_size = base.cam.node().getLens().getFilmSize()
        x = film_size.getX() / 2.0
        z = film_size.getY() / 2.0
        vertex.addData3f(-x, 10000, z)
        vertex.addData3f(x, 10000, z)
        vertex.addData3f(-x, 10000, -z)
        vertex.addData3f(x, 10000, -z)
        color.addData4f(*background_data.color1 + (1.0,))
        color.addData4f(*background_data.color1 + (1.0,))
        color.addData4f(*background_data.color2 + (1.0,))
        color.addData4f(*background_data.color2 + (1.0,))
        primitive.addNextVertices(4)
        primitive.closePrimitive()
        geom = Geom(vdata)
        geom.addPrimitive(primitive)
        node = GeomNode('gnode')
        node.addGeom(geom)
        self.node_path = base.camera.attachNewNode(node)


class Texture(object):
    def __init__(self):
        self.texture = None

    def read(self, texture_data):
        from pandac.PandaModules import PNMImage
        from pandac.PandaModules import Texture as P3DTexture
        tex_pnm = PNMImage(256, 1024)
        for y in range(1024):
            for x in range(256):
                pal_i = texture_data.image[y][x]
                gray = pal_i / 15.0
                tex_pnm.setXel(x, y, gray)
        self.texture = P3DTexture()
        self.texture.load(tex_pnm)

    def make_blank(self):
        from pandac.PandaModules import PNMImage
        from pandac.PandaModules import Texture as P3DTexture
        tex_pnm = PNMImage(2, 2)
        tex_pnm.fill(1, 1, 1)
        self.texture = P3DTexture()
        self.texture.load(tex_pnm)


class World(object):
    def __init__(self, parent):
        self.parent = parent
        self.map = Map()
        self.node_path = None
        self.polygons = None
        self.dir_lights = None
        self.amb_light = None
        self.init_camera()

    def read(self):
        self.node_path = render.attachNewNode('world')
        self.node_path_mesh = self.node_path.attachNewNode('mesh')
        self.node_path_ui = self.node_path.attachNewNode('ui')
        self.map.read()
        self.get_texture()
        self.get_polygons()
        self.set_camera_zoom()
        self.get_dir_lights()
        self.get_amb_light()
        self.get_background()

    def init_camera(self):
        from pandac.PandaModules import OrthographicLens
        lens = OrthographicLens()
        # TODO: Find real aspect ratio
        lens.setAspectRatio(4.0/3.0)
        lens.setNear(-32768)
        lens.setFar(131072)
        base.cam.node().setLens(lens)

    def set_camera_zoom(self):
        base.cam.node().getLens().setFilmSize(self.map.hypotenuse)

    def set_camera_angle(self, heading, azimuth):
        from math import sin, cos, pi
        size_x = abs(self.map.extents[1][0] - self.map.extents[0][0])
        size_y = abs(self.map.extents[1][1] - self.map.extents[0][1])
        size_z = abs(self.map.extents[1][2] - self.map.extents[0][2])
        center_x = size_x / 2.0
        center_y = size_y / 3.0
        center_z = size_z / 2.0
        (center_x, center_y, center_z) = coords_to_panda(center_x, -center_y, center_z)
        y = cos(pi * azimuth / 180) * self.map.hypotenuse
        z = sin(pi * azimuth / 180) * self.map.hypotenuse
        x = cos(pi * heading / 180) * y
        y = sin(pi * heading / 180) * y
        base.camera.setPos(center_x + x, center_y + y, center_z + z)
        from pandac.PandaModules import Point3, Vec3
        base.camera.lookAt(Point3(center_x, center_y, center_z))

    def spin_camera(self, task):
        heading = task.time * 30
        azimuth = 30
        self.set_camera_angle(heading, azimuth)
        return task.cont

    def get_texture(self):
        texture = Texture()
        texture.read(self.map.get_texture())
        self.texture = texture
        self.node_path_mesh.setTexture(self.texture.texture)

    def get_polygons(self):
        polygons = []
        for i, poly_data in enumerate(self.map.get_polygons()):
            polygon = Polygon(self)
            polygon.init_node_path(poly_data)
            polygon.node_path.setTag('polygon_i', str(i))
            polygons.append(polygon)
        self.polygons = polygons

    def get_dir_lights(self):
        dir_lights = []
        for dir_light_data in self.map.get_dir_lights():
            dir_light = Directional_Light(self)
            dir_lights.append(dir_light)
            dir_light.init_node_path(dir_light_data)
        self.dir_lights = dir_lights

    def get_amb_light(self):
        amb_light_data = self.map.get_amb_light()
        amb_light = Ambient_Light(self)
        amb_light.init_node_path(amb_light_data)
        self.amb_light = amb_light

    def get_background(self):
        background_data = self.map.get_background()
        background = Background(self)
        background.init_node_path(background_data)
        self.background = background

    def read_gns(self, gns_path):
        if gns_path is None:
            gns_path = self.parent.file_dialog()
        assert gns_path is not None, 'No GNS file chosen. Exiting.'
        self.map.gns.read(gns_path)
        self.map.set_situation(0)

    def next_gns(self):
        import os
        gns_path = self.map.gns.file_path
        gns_dir = os.path.dirname(gns_path)
        gns_name = os.path.basename(gns_path)
        files = []
        for file_name in os.listdir(gns_dir):
            if file_name[-4:] in ['.gns', '.GNS']:
                files.append(file_name)
        files.sort()
        found = False
        for i, file_name in enumerate(files):
            if file_name == gns_name:
                found = True
                break
        if found:
            new_i = (i + 1) % len(files)
            new_file_name = files[new_i]
            new_gns_path = os.path.join(gns_dir, new_file_name)
            self.read_gns(new_gns_path)
            self.read()

    def next_situation(self):
        sit = self.map.situation + 1
        self.map.set_situation(sit)
        self.read()

    def prev_situation(self):
        sit = self.map.situation - 1
        self.map.set_situation(sit)
        self.read()


